"""
nexon.ext.slash_utils
~~~~~~~~~~~~~~~~~~~~~~~~
An extension module to provide utilities for slash commands.
:copyright: (c) 2022-present Nextcord Developers
:license: MIT, see LICENSE for more details.
"""

from .slash import describe
